@@ -1,6 +1,7 @@
 /*
  * Copyright (c) 2015 Los Alamos Nat. Security, LLC. All rights reserved.
  * Copyright (c) 2016 Cisco Systems, Inc. All rights reserved.
+ * Copyright (c) 2018 Amazon.com, Inc. or its affiliates. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -44,8 +45,10 @@
 
 #include <pthread.h>
 
-#include "unix/osd.h"
+#include <ifaddrs.h>
 
+#include "unix/osd.h"
+#include "rdma/fi_errno.h"
 #include "config.h"
 
 #define pthread_yield pthread_yield_np
@@ -67,6 +70,51 @@ static inline int ofi_shm_remap(struct util_shm *shm, size_t newsize, void **map
 	return -1;
 }
 
+static inline ssize_t ofi_get_hugepage_size(void)
+{
+	return -FI_ENOSYS;
+}
+
+static inline int ofi_alloc_hugepage_buf(void **memptr, size_t size)
+{
+	return -FI_ENOSYS;
+}
+
+static inline int ofi_free_hugepage_buf(void *memptr, size_t size)
+{
+	return -FI_ENOSYS;
+}
+
+static inline size_t ofi_ifaddr_get_speed(struct ifaddrs *ifa)
+{
+	return 0;
+}
+
+static inline int ofi_hugepage_enabled(void)
+{
+	return 0;
+}
+
+static inline ssize_t ofi_process_vm_readv(pid_t pid,
+			const struct iovec *local_iov,
+			unsigned long liovcnt,
+			const struct iovec *remote_iov,
+			unsigned long riovcnt,
+			unsigned long flags)
+{
+	return -FI_ENOSYS;
+}
+
+static inline size_t ofi_process_vm_writev(pid_t pid,
+			 const struct iovec *local_iov,
+			 unsigned long liovcnt,
+			 const struct iovec *remote_iov,
+			 unsigned long riovcnt,
+			 unsigned long flags)
+{
+	return -FI_ENOSYS;
+}
+
 #ifdef __cplusplus
 }
 #endif
