@@ -38,7 +38,6 @@
 #include "netdir.h"
 #include "netdir_ov.h"
 #include "netdir_cq.h"
-#include "netdir_err.h"
 #include "netdir_log.h"
 #include "netdir_iface.h"
 #include "netdir_unexp.h"
@@ -46,8 +45,8 @@
 #include "rdma/fabric.h"
 #include "rdma/fi_endpoint.h"
 
-#include "fi.h"
-#include "fi_util.h"
+#include "ofi.h"
+#include "ofi_util.h"
 
 static ssize_t ofi_nd_ep_recv(struct fid_ep *ep, void *buf, size_t len,
 			      void *desc, fi_addr_t src_addr, void *context);
@@ -161,12 +160,12 @@ static int ofi_nd_ep_sendmsg_inline(struct nd_ep *ep,
 			ND2_SGE sge_def = {
 				.Buffer = msg->msg_iov[i].iov_base,
 				.BufferLength = (ULONG)msg->msg_iov[i].iov_len,
-				.MemoryRegionToken = (UINT32)(msg->desc[i])
+				.MemoryRegionToken = (UINT32)(uintptr_t)msg->desc[i]
 			};
 			sge_entry->entries[i + 1] = sge_def;
 		}
 
-		sge_entry->count = msg->iov_count + 1;
+		sge_entry->count = (ULONG)msg->iov_count + 1;
 	}
 
 	nd_send_entry *send_entry = ofi_nd_buf_alloc_nd_send_entry();
@@ -308,7 +307,7 @@ static int ofi_nd_ep_sendmsg_large(struct nd_ep *ep,
 		},
 		{
 			.Buffer = entry->notify_buf->location,
-			.BufferLength = (ULONG)sizeof(*entry->notify_buf->location) * msg->iov_count,
+			.BufferLength = (ULONG)(sizeof(*entry->notify_buf->location) * msg->iov_count),
 			.MemoryRegionToken = entry->notify_buf->token
 		}
 	};
