@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2017 Intel Corporation. All rights reserved.
+ * Copyright (c) 2013-2019 Intel Corporation. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -32,109 +32,73 @@
 
 #include "psmx2.h"
 
-#define BIT(i)		(1ULL << i)
-#define BITMAP_SIZE	(PSMX2_MAX_VL + 1)
-
-static inline void bitmap_set(uint64_t *map, unsigned id)
-{
-	int i, j;
-
-	i = id / sizeof(uint64_t);
-	j = id % sizeof(uint64_t);
-
-	map[i] |= BIT(j);
-}
-
-static inline void bitmap_clear(uint64_t *map, unsigned id)
-{
-	int i, j;
-
-	i = id / sizeof(uint64_t);
-	j = id % sizeof(uint64_t);
-
-	map[i] &= ~BIT(j);
-}
-
-static inline int bitmap_test(uint64_t *map, unsigned id)
-{
-	int i, j;
-
-	i = id / sizeof(uint64_t);
-	j = id % sizeof(uint64_t);
-
-	return !!(map[i] & BIT(j));
-}
-
-static void psmx2_free_vlane(struct psmx2_fid_domain *domain, uint8_t vl)
-{
-	psmx2_lock(&domain->vl_lock, 1);
-	bitmap_clear(domain->vl_map, vl);
-	psmx2_unlock(&domain->vl_lock, 1);
-}
-
-static int psmx2_alloc_vlane(struct psmx2_fid_domain *domain, uint8_t *vl)
-{
-	int i;
-	int id;
-
-	psmx2_lock(&domain->vl_lock, 1);
-	for (i = 0; i < BITMAP_SIZE; i++) {
-		id = (domain->vl_alloc + i) % BITMAP_SIZE;
-		if (bitmap_test(domain->vl_map, id) == 0) {
-			bitmap_set(domain->vl_map, id);
-			domain->vl_alloc = id + 1;
-			break;
-		}
-	}
-	psmx2_unlock(&domain->vl_lock, 1);
-
-	if (i >= BITMAP_SIZE)
-		return -FI_ENOSPC;
-
-	*vl = (uint8_t)id;
-	return 0;
-}
+#define PSMX2_EP_SET_TAGGED_OPS(suffix, msg_suffix)				\
+	do {									\
+		if (!send_completion && !recv_completion) {			\
+			ep->ep.tagged = &psmx2_tagged_ops_no_event##suffix;	\
+			FI_INFO(&psmx2_prov, FI_LOG_EP_DATA,			\
+				"tagged ops optimized for op_flags=0 "		\
+				"and event suppression "			\
+				msg_suffix					\
+				"\n");						\
+		} else if (!send_completion) {					\
+			ep->ep.tagged = &psmx2_tagged_ops_no_send_event##suffix;\
+			FI_INFO(&psmx2_prov, FI_LOG_EP_DATA,			\
+				"tagged ops optimized for op_flags=0 "		\
+				"and send event suppression "			\
+				msg_suffix					\
+				"\n");						\
+		} else if (!recv_completion) {					\
+			ep->ep.tagged = &psmx2_tagged_ops_no_recv_event##suffix;\
+			FI_INFO(&psmx2_prov, FI_LOG_EP_DATA,			\
+				"tagged ops optimized for op_flags=0 "		\
+				"and recv event suppression "			\
+				msg_suffix					\
+				"\n");						\
+		} else {							\
+			ep->ep.tagged = &psmx2_tagged_ops_no_flag##suffix;	\
+			FI_INFO(&psmx2_prov, FI_LOG_EP_DATA,			\
+				"tagged ops optimized for op_flags=0 "		\
+				msg_suffix					\
+				"\n");						\
+		}								\
+	} while (0)
 
 static void psmx2_ep_optimize_ops(struct psmx2_fid_ep *ep)
 {
+	int send_completion;
+	int recv_completion;
+	uint64_t mask;
+
+	mask = PSMX2_OP_FLAGS &
+	       ~(FI_INJECT_COMPLETE | FI_TRANSMIT_COMPLETE | FI_DELIVERY_COMPLETE);
+
 	if (ep->ep.tagged) {
-		if (ep->tx_flags | ep->rx_flags) {
+		if (ep->tx_flags & mask & ~FI_COMPLETION || ep->rx_flags & mask & ~FI_COMPLETION) {
 			ep->ep.tagged = &psmx2_tagged_ops;
 			FI_INFO(&psmx2_prov, FI_LOG_EP_DATA,
 				"generic tagged ops.\n");
-		} else if (ep->send_selective_completion && ep->recv_selective_completion) {
-			if (ep->av && ep->av->type == FI_AV_TABLE)
-				ep->ep.tagged = &psmx2_tagged_ops_no_event_av_table;
-			else
-				ep->ep.tagged = &psmx2_tagged_ops_no_event_av_map;
-			FI_INFO(&psmx2_prov, FI_LOG_EP_DATA,
-				"tagged ops optimized for op_flags=0 and event suppression\n");
-		} else if (ep->send_selective_completion) {
-			if (ep->av && ep->av->type == FI_AV_TABLE)
-				ep->ep.tagged = &psmx2_tagged_ops_no_send_event_av_table;
-			else
-				ep->ep.tagged = &psmx2_tagged_ops_no_send_event_av_map;
-			FI_INFO(&psmx2_prov, FI_LOG_EP_DATA,
-				"tagged ops optimized for op_flags=0 and send event suppression\n");
-		} else if (ep->recv_selective_completion) {
-			if (ep->av && ep->av->type == FI_AV_TABLE)
-				ep->ep.tagged = &psmx2_tagged_ops_no_recv_event_av_table;
-			else
-				ep->ep.tagged = &psmx2_tagged_ops_no_recv_event_av_map;
-			FI_INFO(&psmx2_prov, FI_LOG_EP_DATA,
-				"tagged ops optimized for op_flags=0 and recv event suppression\n");
 		} else {
-			if (ep->av && ep->av->type == FI_AV_TABLE)
-				ep->ep.tagged = &psmx2_tagged_ops_no_flag_av_table;
-			else
-				ep->ep.tagged = &psmx2_tagged_ops_no_flag_av_map;
-			FI_INFO(&psmx2_prov, FI_LOG_EP_DATA,
-				"tagged ops optimized for op_flags=0\n");
+			send_completion = !ep->send_selective_completion || ep->tx_flags & FI_COMPLETION;
+			recv_completion = !ep->recv_selective_completion || ep->rx_flags & FI_COMPLETION;
+
+			if (ep->av && ep->av->type == FI_AV_MAP) {
+				if (ep->caps & FI_DIRECTED_RECV)
+					PSMX2_EP_SET_TAGGED_OPS(_directed_av_map, "and directed receive and av map");
+				else
+					PSMX2_EP_SET_TAGGED_OPS(_undirected_av_map, "and av map");
+			} else {
+				if (ep->caps & FI_DIRECTED_RECV)
+					PSMX2_EP_SET_TAGGED_OPS(_directed, "and directed receive");
+				else
+					PSMX2_EP_SET_TAGGED_OPS(_undirected, "");
+			}
 		}
 	}
 }
 
-static ssize_t psmx2_ep_cancel(fid_t fid, void *context)
+DIRECT_FN
+STATIC ssize_t psmx2_ep_cancel(fid_t fid, void *context)
 {
 	struct psmx2_fid_ep *ep;
 	psm2_mq_status2_t status;
@@ -144,11 +108,8 @@ static ssize_t psmx2_ep_cancel(fid_t fid, void *context)
 	int err;
 
 	ep = container_of(fid, struct psmx2_fid_ep, ep.fid);
-	if (!ep->domain)
-		return -FI_EBADF;
-
-	if (!fi_context)
-		return -FI_EINVAL;
+	assert(ep->domain);
+	assert(fi_context);
 
 	switch (PSMX2_CTXT_TYPE(fi_context)) {
 	case PSMX2_TRECV_CONTEXT:
@@ -186,7 +147,8 @@ static ssize_t psmx2_ep_cancel(fid_t fid, void *context)
 	return psmx2_errno(err);
 }
 
-static int psmx2_ep_getopt(fid_t fid, int level, int optname,
+DIRECT_FN
+STATIC int psmx2_ep_getopt(fid_t fid, int level, int optname,
 			   void *optval, size_t *optlen)
 {
 	struct psmx2_fid_ep *ep;
@@ -209,7 +171,8 @@ static int psmx2_ep_getopt(fid_t fid, int level, int optname,
 	return 0;
 }
 
-static int psmx2_ep_setopt(fid_t fid, int level, int optname,
+DIRECT_FN
+STATIC int psmx2_ep_setopt(fid_t fid, int level, int optname,
 			   const void *optval, size_t optlen)
 {
 	struct psmx2_fid_ep *ep;
@@ -233,18 +196,8 @@ static int psmx2_ep_setopt(fid_t fid, int level, int optname,
 
 static void psmx2_ep_close_internal(struct psmx2_fid_ep *ep)
 {
-	struct slist_entry *entry;
-	struct psmx2_context *item;
-
 	psmx2_domain_release(ep->domain);
-
-	while (!slist_empty(&ep->free_context_list)) {
-		entry = slist_remove_head(&ep->free_context_list);
-		item = container_of(entry, struct psmx2_context, list_entry);
-		free(item);
-	}
-
-	fastlock_destroy(&ep->context_lock);
+	PSMX2_EP_FINI_OP_CONTEXT(ep);
 	free(ep);
 }
 
@@ -252,6 +205,7 @@ static int psmx2_ep_close(fid_t fid)
 {
 	struct psmx2_fid_ep *ep;
 	struct psmx2_ep_name ep_name;
+	int usage_flags = 0;
 
 	ep = container_of(fid, struct psmx2_fid_ep, ep.fid);
 
@@ -263,20 +217,56 @@ static int psmx2_ep_close(fid_t fid)
 	if (ofi_atomic_get32(&ep->ref))
 		return -FI_EBUSY;
 
-	ep_name.epid = ep->trx_ctxt->psm2_epid;
-	ep_name.vlane = ep->vlane;
+	if (ep->stx)
+		ofi_atomic_dec32(&ep->stx->ref);
 
-	ofi_ns_del_local_name(&ep->domain->fabric->name_server,
-			      &ep->service, &ep_name);
+	if (ep->tx && !ep->stx)
+		usage_flags |= PSMX2_TX;
 
-	ep->domain->eps[ep->vlane] = NULL;
-	psmx2_free_vlane(ep->domain, ep->vlane);
+	if (ep->rx) {
+		usage_flags |= PSMX2_RX;
+		ep_name.epid = ep->rx->psm2_epid;
 
+		ofi_ns_del_local_name(&ep->domain->fabric->name_server,
+				      &ep->service, &ep_name);
+	}
+
+	psmx2_trx_ctxt_free(ep->rx, usage_flags);
 	psmx2_ep_close_internal(ep);
 	return 0;
 }
 
-static int psmx2_ep_bind(struct fid *fid, struct fid *bfid, uint64_t flags)
+static int psmx2_poll_ctxt_match(struct slist_entry *entry, const void *arg)
+{
+	struct psmx2_poll_ctxt *poll_ctxt;
+
+	poll_ctxt = container_of(entry, struct psmx2_poll_ctxt, list_entry);
+	return (poll_ctxt->trx_ctxt == arg);
+}
+
+static int psmx2_add_poll_ctxt(struct slist *list, struct psmx2_trx_ctxt *trx_ctxt)
+{
+	struct psmx2_poll_ctxt *item;
+
+	if (!trx_ctxt)
+		return 0;
+
+	if (!slist_empty(list) &&
+	    slist_find_first_match(list, psmx2_poll_ctxt_match, trx_ctxt))
+		return 0;
+
+	item = calloc(1, sizeof(*item));
+	if (!item)
+		return -FI_ENOMEM;
+
+	ofi_atomic_inc32(&trx_ctxt->poll_refcnt);
+	item->trx_ctxt = trx_ctxt;
+	slist_insert_tail(&item->list_entry, list);
+	return 0;
+}
+
+DIRECT_FN
+STATIC int psmx2_ep_bind(struct fid *fid, struct fid *bfid, uint64_t flags)
 {
 	struct psmx2_fid_ep *ep;
 	struct psmx2_fid_av *av;
@@ -298,19 +288,18 @@ static int psmx2_ep_bind(struct fid *fid, struct fid *bfid, uint64_t flags)
 		cq = container_of(bfid, struct psmx2_fid_cq, cq.fid);
 		if (ep->domain != cq->domain)
 			return -FI_EINVAL;
-		if (cq->trx_ctxt == PSMX2_ALL_TRX_CTXT) {
-			/* do nothing */
-		} else if (!cq->trx_ctxt) {
-			cq->trx_ctxt = ep->trx_ctxt;
-		} else if (cq->trx_ctxt != ep->trx_ctxt) {
-			return -FI_EINVAL;
-		}
 		if (flags & FI_SEND) {
+			err = psmx2_add_poll_ctxt(&cq->poll_list, ep->tx);
+			if (err)
+				return err;
 			ep->send_cq = cq;
 			if (flags & FI_SELECTIVE_COMPLETION)
 				ep->send_selective_completion = 1;
 		}
 		if (flags & FI_RECV) {
+			err = psmx2_add_poll_ctxt(&cq->poll_list, ep->rx);
+			if (err)
+				return err;
 			ep->recv_cq = cq;
 			if (flags & FI_SELECTIVE_COMPLETION)
 				ep->recv_selective_completion = 1;
@@ -322,12 +311,15 @@ static int psmx2_ep_bind(struct fid *fid, struct fid *bfid, uint64_t flags)
 		cntr = container_of(bfid, struct psmx2_fid_cntr, cntr.fid);
 		if (ep->domain != cntr->domain)
 			return -FI_EINVAL;
-		if (cntr->trx_ctxt == PSMX2_ALL_TRX_CTXT) {
-			/* do nothing */
-		} else if (!cntr->trx_ctxt) {
-			cntr->trx_ctxt = ep->trx_ctxt;
-		} else if (cntr->trx_ctxt != ep->trx_ctxt) {
-			return -FI_EINVAL;
+		if (flags & (FI_SEND | FI_WRITE | FI_READ)) {
+			err = psmx2_add_poll_ctxt(&cntr->poll_list, ep->tx);
+			if (err)
+				return err;
+		}
+		if (flags & (FI_RECV | FI_REMOTE_WRITE | FI_REMOTE_READ)) {
+			err = psmx2_add_poll_ctxt(&cntr->poll_list, ep->rx);
+			if (err)
+				return err;
 		}
 		if (flags & FI_SEND)
 			ep->send_cntr = cntr;
@@ -350,6 +342,10 @@ static int psmx2_ep_bind(struct fid *fid, struct fid *bfid, uint64_t flags)
 			return -FI_EINVAL;
 		ep->av = av;
 		psmx2_ep_optimize_ops(ep);
+		if (ep->tx)
+			psmx2_av_add_trx_ctxt(av, ep->tx);
+		if (ep->rx && ep->rx != ep->tx)
+			psmx2_av_add_trx_ctxt(av, ep->rx);
 		break;
 
 	case FI_CLASS_MR:
@@ -361,10 +357,23 @@ static int psmx2_ep_bind(struct fid *fid, struct fid *bfid, uint64_t flags)
 		break;
 
 	case FI_CLASS_STX_CTX:
-		stx = container_of(bfid,
-				   struct psmx2_fid_stx, stx.fid);
+		stx = container_of(bfid, struct psmx2_fid_stx, stx.fid);
 		if (ep->domain != stx->domain)
 			return -FI_EINVAL;
+		if (ep->tx || ep->stx)
+			return -FI_EINVAL;
+		ep->tx = stx->tx;
+		ep->stx = stx;
+		err = psmx2_domain_enable_ep(ep->domain, ep);
+		if (err)
+			return err;
+#if HAVE_PSM2_MQ_FP_MSG
+		if (ep->caps & FI_TRIGGER)
+#else
+		if (ep->caps & (FI_RMA | FI_TRIGGER))
+#endif
+			stx->tx->am_progress = 1;
+		ofi_atomic_inc32(&stx->ref);
 		break;
 
 	default:
@@ -406,7 +415,8 @@ static inline int psmx2_ep_get_flags(struct psmx2_fid_ep *ep, uint64_t *flags)
 	return 0;
 }
 
-static int psmx2_ep_control(fid_t fid, int command, void *arg)
+DIRECT_FN
+STATIC int psmx2_ep_control(fid_t fid, int command, void *arg)
 {
 	struct fi_alias *alias;
 	struct psmx2_fid_ep *ep, *new_ep;
@@ -458,7 +468,8 @@ static int psmx2_ep_control(fid_t fid, int command, void *arg)
 	return 0;
 }
 
-static ssize_t psmx2_rx_size_left(struct fid_ep *ep)
+DIRECT_FN
+STATIC ssize_t psmx2_rx_size_left(struct fid_ep *ep)
 {
 	struct psmx2_fid_ep *ep_priv;
 
@@ -469,7 +480,8 @@ static ssize_t psmx2_rx_size_left(struct fid_ep *ep)
 		return -FI_EOPBADSTATE;
 }
 
-static ssize_t psmx2_tx_size_left(struct fid_ep *ep)
+DIRECT_FN
+STATIC ssize_t psmx2_tx_size_left(struct fid_ep *ep)
 {
 	struct psmx2_fid_ep *ep_priv;
 
@@ -502,13 +514,12 @@ static struct fi_ops_ep psmx2_ep_ops = {
 int psmx2_ep_open_internal(struct psmx2_fid_domain *domain_priv,
 			   struct fi_info *info,
 			   struct psmx2_fid_ep **ep_out, void *context,
-			   struct psmx2_trx_ctxt *trx_ctxt, int vlane)
+			   struct psmx2_trx_ctxt *trx_ctxt,
+			   int usage_flags)
 {
 	struct psmx2_fid_ep *ep_priv;
-	struct psmx2_context *item;
 	uint64_t ep_cap;
 	int err = -FI_EINVAL;
-	int i;
 
 	if (info)
 		ep_cap = info->caps;
@@ -533,10 +544,6 @@ int psmx2_ep_open_internal(struct psmx2_fid_domain *domain_priv,
 		}
 	}
 
-	err = psmx2_domain_check_features(domain_priv, ep_cap);
-	if (err)
-		goto errout;
-
 	ep_priv = (struct psmx2_fid_ep *) calloc(1, sizeof *ep_priv);
 	if (!ep_priv) {
 		err = -FI_ENOMEM;
@@ -549,14 +556,19 @@ int psmx2_ep_open_internal(struct psmx2_fid_domain *domain_priv,
 	ep_priv->ep.ops = &psmx2_ep_ops;
 	ep_priv->ep.cm = &psmx2_cm_ops;
 	ep_priv->domain = domain_priv;
-	ep_priv->trx_ctxt = trx_ctxt;
-	ep_priv->vlane = vlane;
+	if (usage_flags & PSMX2_RX) {
+		ep_priv->rx = trx_ctxt;
+		if (trx_ctxt)
+			trx_ctxt->ep = ep_priv; /* only used by RMA target */
+	}
+	if (usage_flags & PSMX2_TX)
+		ep_priv->tx = trx_ctxt;
 	ofi_atomic_initialize32(&ep_priv->ref, 0);
 
 	PSMX2_CTXT_TYPE(&ep_priv->nocomp_send_context) = PSMX2_NOCOMP_SEND_CONTEXT;
 	PSMX2_CTXT_EP(&ep_priv->nocomp_send_context) = ep_priv;
-	PSMX2_CTXT_TYPE(&ep_priv->nocomp_recv_context) = PSMX2_NOCOMP_RECV_CONTEXT;
-	PSMX2_CTXT_EP(&ep_priv->nocomp_recv_context) = ep_priv;
+	PSMX2_CTXT_TYPE(&ep_priv->nocomp_tsend_context) = PSMX2_NOCOMP_TSEND_CONTEXT;
+	PSMX2_CTXT_EP(&ep_priv->nocomp_tsend_context) = ep_priv;
 
 	if (ep_cap & FI_TAGGED)
 		ep_priv->ep.tagged = &psmx2_tagged_ops;
@@ -584,19 +596,13 @@ int psmx2_ep_open_internal(struct psmx2_fid_domain *domain_priv,
 
 	psmx2_ep_optimize_ops(ep_priv);
 
-	slist_init(&ep_priv->free_context_list);
-	fastlock_init(&ep_priv->context_lock);
-
-#define PSMX2_FREE_CONTEXT_LIST_SIZE	64
-	for (i = 0; i < PSMX2_FREE_CONTEXT_LIST_SIZE; i++) {
-		item = calloc(1, sizeof(*item));
-		if (!item) {
-			FI_WARN(&psmx2_prov, FI_LOG_EP_CTRL, "out of memory.\n");
-			exit(-1);
-		}
-		slist_insert_tail(&item->list_entry, &ep_priv->free_context_list);
-	}
-
+	PSMX2_EP_INIT_OP_CONTEXT(ep_priv);
+#if HAVE_PSM2_MQ_FP_MSG
+	if ((ep_cap & FI_TRIGGER) && trx_ctxt)
+#else
+	if ((ep_cap & (FI_RMA | FI_TRIGGER)) && trx_ctxt)
+#endif
+		trx_ctxt->am_progress = 1;
 
 	*ep_out = ep_priv;
 	return 0;
@@ -608,6 +614,7 @@ errout:
 	return err;
 }
 
+DIRECT_FN
 int psmx2_ep_open(struct fid_domain *domain, struct fi_info *info,
 		  struct fid_ep **ep, void *context)
 {
@@ -615,61 +622,86 @@ int psmx2_ep_open(struct fid_domain *domain, struct fi_info *info,
 	struct psmx2_fid_ep *ep_priv;
 	struct psmx2_ep_name ep_name;
 	struct psmx2_ep_name *src_addr;
-	uint8_t vlane;
+	struct psmx2_trx_ctxt *trx_ctxt = NULL;
 	int err = -FI_EINVAL;
+	int usage_flags = PSMX2_TX_RX;
 
 	domain_priv = container_of(domain, struct psmx2_fid_domain,
 				   util_domain.domain_fid.fid);
 	if (!domain_priv)
 		goto errout;
 
-	err = psmx2_alloc_vlane(domain_priv, &vlane);
-	if (err)
-		goto errout;
+	if (info && info->ep_attr &&
+	    info->ep_attr->rx_ctx_cnt == FI_SHARED_CONTEXT)
+		return  -FI_ENOSYS;
 
-	err = psmx2_ep_open_internal(domain_priv, info, &ep_priv, context,
-				     domain_priv->base_trx_ctxt, vlane);
-	if (err) {
-		psmx2_free_vlane(domain_priv, vlane);
-		goto errout;
+	if (info && info->ep_attr &&
+	    info->ep_attr->tx_ctx_cnt == FI_SHARED_CONTEXT)
+		usage_flags &= ~PSMX2_TX;
+
+	if (info && !ofi_send_allowed(info->caps) &&
+	    !ofi_rma_initiate_allowed(info->caps))
+		usage_flags &= ~PSMX2_TX;
+
+	if (info && !ofi_recv_allowed(info->caps) &&
+	    !ofi_rma_target_allowed(info->caps))
+		usage_flags &= ~PSMX2_RX;
+
+	src_addr = NULL;
+	if (info && info->src_addr) {
+		if (info->addr_format == FI_ADDR_STR)
+			src_addr = psmx2_string_to_ep_name(info->src_addr);
+		else
+			src_addr = info->src_addr;
 	}
 
-	domain_priv->eps[ep_priv->vlane] = ep_priv;
+	if (usage_flags) {
+		trx_ctxt = psmx2_trx_ctxt_alloc(domain_priv, src_addr, -1,
+						usage_flags);
+		if (!trx_ctxt)
+			goto errout;
+	} else {
+		FI_INFO(&psmx2_prov, FI_LOG_EP_CTRL,
+			"Tx only endpoint with STX context.\n");
+	}
+
+	err = psmx2_ep_open_internal(domain_priv, info, &ep_priv, context,
+				     trx_ctxt, usage_flags);
+	if (err)
+		goto errout_free_ctxt;
 
 	ep_priv->type = PSMX2_EP_REGULAR;
 	ep_priv->service = PSMX2_ANY_SERVICE;
-	if (info && info->src_addr) {
-		if (info->addr_format == FI_ADDR_STR) {
-			src_addr = psmx2_string_to_ep_name(info->src_addr);
-			if (src_addr) {
-				ep_priv->service = src_addr->service;
-				free(src_addr);
-			}
-		} else {
-			ep_priv->service = ((struct psmx2_ep_name *)info->src_addr)->service;
-		}
+	if (src_addr) {
+		ep_priv->service = src_addr->service;
+		if (info->addr_format == FI_ADDR_STR)
+			free(src_addr);
 	}
 
 	if (ep_priv->service == PSMX2_ANY_SERVICE)
 		ep_priv->service = ((getpid() & 0x7FFF) << 16) +
 				   ((uintptr_t)ep_priv & 0xFFFF);
 
-	ep_name.epid = domain_priv->base_trx_ctxt->psm2_epid;
-	ep_name.vlane = ep_priv->vlane;
-	ep_name.type = ep_priv->type;
+	if (usage_flags) {
+		ep_name.epid = trx_ctxt->psm2_epid;
+		ep_name.type = ep_priv->type;
 
-	ofi_ns_add_local_name(&domain_priv->fabric->name_server,
-			      &ep_priv->service, &ep_name);
+		ofi_ns_add_local_name(&domain_priv->fabric->name_server,
+				      &ep_priv->service, &ep_name);
+	}
 
 	*ep = &ep_priv->ep;
 	return 0;
 
+errout_free_ctxt:
+	psmx2_trx_ctxt_free(trx_ctxt, usage_flags);
+
 errout:
 	return err;
 }
 
-/* STX support is essentially no-op since PSM supports only one send/recv
- * context and thus always works in shared context mode.
+/*
+ * Shared tx context
  */
 
 static int psmx2_stx_close(fid_t fid)
@@ -677,9 +709,13 @@ static int psmx2_stx_close(fid_t fid)
 	struct psmx2_fid_stx *stx;
 
 	stx = container_of(fid, struct psmx2_fid_stx, stx.fid);
+
+	if (ofi_atomic_get32(&stx->ref))
+		return -FI_EBUSY;
+
+	psmx2_trx_ctxt_free(stx->tx, PSMX2_TX);
 	psmx2_domain_release(stx->domain);
 	free(stx);
-
 	return 0;
 }
 
@@ -688,69 +724,63 @@ static struct fi_ops psmx2_fi_ops_stx = {
 	.close = psmx2_stx_close,
 	.bind = fi_no_bind,
 	.control = fi_no_control,
+	.ops_open = fi_no_ops_open,
 };
 
+static struct fi_ops_ep psmx2_stx_ops = {
+	.size = sizeof(struct fi_ops_ep),
+	.cancel = fi_no_cancel,
+	.getopt = fi_no_getopt,
+	.setopt = fi_no_setopt,
+	.tx_ctx = fi_no_tx_ctx,
+	.rx_ctx = fi_no_rx_ctx,
+	.rx_size_left = fi_no_rx_size_left,
+	.tx_size_left = fi_no_tx_size_left,
+};
+
+DIRECT_FN
 int psmx2_stx_ctx(struct fid_domain *domain, struct fi_tx_attr *attr,
-		 struct fid_stx **stx, void *context)
+		  struct fid_stx **stx, void *context)
 {
 	struct psmx2_fid_domain *domain_priv;
+	struct psmx2_trx_ctxt *trx_ctxt;
 	struct psmx2_fid_stx *stx_priv;
-
-	FI_INFO(&psmx2_prov, FI_LOG_EP_DATA, "\n");
+	int err = -FI_EINVAL;
 
 	domain_priv = container_of(domain, struct psmx2_fid_domain,
 				   util_domain.domain_fid.fid);
+	if (!domain_priv)
+		goto errout;
 
 	stx_priv = (struct psmx2_fid_stx *) calloc(1, sizeof *stx_priv);
-	if (!stx_priv)
-		return -FI_ENOMEM;
+	if (!stx_priv) {
+		err = -FI_ENOMEM;
+		goto errout;
+	}
+
+	trx_ctxt = psmx2_trx_ctxt_alloc(domain_priv, NULL, -1, PSMX2_TX);
+	if (!trx_ctxt) {
+		err = -FI_ENOMEM;
+		goto errout_free_stx;
+	}
 
 	psmx2_domain_acquire(domain_priv);
-
 	stx_priv->stx.fid.fclass = FI_CLASS_STX_CTX;
 	stx_priv->stx.fid.context = context;
 	stx_priv->stx.fid.ops = &psmx2_fi_ops_stx;
+	stx_priv->stx.ops = &psmx2_stx_ops;
 	stx_priv->domain = domain_priv;
+	stx_priv->tx = trx_ctxt;
+	ofi_atomic_initialize32(&stx_priv->ref, 0);
 
 	*stx = &stx_priv->stx;
 	return 0;
-}
 
-/* Private op context pool management */
+errout_free_stx:
+	free(stx_priv);
 
-struct fi_context *psmx2_ep_get_op_context(struct psmx2_fid_ep *ep)
-{
-	struct psmx2_context *context;
-
-	psmx2_lock(&ep->context_lock, 2);
-	if (!slist_empty(&ep->free_context_list)) {
-		context = container_of(slist_remove_head(&ep->free_context_list),
-				       struct psmx2_context, list_entry);
-		psmx2_unlock(&ep->context_lock, 2);
-		return &context->fi_context;
-	}
-	psmx2_unlock(&ep->context_lock, 2);
-
-	context = malloc(sizeof(*context));
-	if (!context)
-		FI_WARN(&psmx2_prov, FI_LOG_EP_DATA, "out of memory.\n");
-
-	return &context->fi_context;
-}
-
-void psmx2_ep_put_op_context(struct psmx2_fid_ep *ep,
-			     struct fi_context *fi_context)
-{
-	struct psmx2_context *context;
-
-	if (PSMX2_CTXT_TYPE(fi_context) != PSMX2_NOCOMP_RECV_CONTEXT_ALLOC)
-		return;
-
-	context = container_of(fi_context, struct psmx2_context, fi_context);
-	context->list_entry.next = NULL;
-	psmx2_lock(&ep->context_lock, 2);
-	slist_insert_tail(&context->list_entry, &ep->free_context_list);
-	psmx2_unlock(&ep->context_lock, 2);
+errout:
+	return err;
 }
 
 /*
@@ -773,26 +803,23 @@ static int psmx2_sep_close(fid_t fid)
 			return -FI_EBUSY;
 	}
 
-	for (i = 0; i < sep->ctxt_cnt; i++) {
-		if (sep->ctxts[i].ep)
-			psmx2_ep_close_internal(sep->ctxts[i].ep);
-
-		psmx2_lock(&sep->domain->trx_ctxt_lock, 1);
-		dlist_remove(&sep->ctxts[i].trx_ctxt->entry);
-		psmx2_unlock(&sep->domain->trx_ctxt_lock, 1);
-		psmx2_trx_ctxt_free(sep->ctxts[i].trx_ctxt);
-	}
-
-	ep_name.epid = sep->domain->base_trx_ctxt->psm2_epid;
+	ep_name.epid = sep->ctxts[0].trx_ctxt->psm2_epid;
 	ep_name.sep_id = sep->id;
 	ep_name.type = sep->type;
 
 	ofi_ns_del_local_name(&sep->domain->fabric->name_server,
 			      &sep->service, &ep_name);
 
-	psmx2_lock(&sep->domain->sep_lock, 1);
+	for (i = 0; i < sep->ctxt_cnt; i++) {
+		psmx2_trx_ctxt_free(sep->ctxts[i].trx_ctxt, PSMX2_TX_RX);
+
+		if (sep->ctxts[i].ep)
+			psmx2_ep_close_internal(sep->ctxts[i].ep);
+	}
+
+	sep->domain->sep_lock_fn(&sep->domain->sep_lock, 1);
 	dlist_remove(&sep->entry);
-	psmx2_unlock(&sep->domain->sep_lock, 1);
+	sep->domain->sep_unlock_fn(&sep->domain->sep_lock, 1);
 
 	psmx2_domain_release(sep->domain);
 	free(sep);
@@ -817,7 +844,8 @@ static int psmx2_sep_control(fid_t fid, int command, void *arg)
 	return 0;
 }
 
-static int psmx2_sep_bind(struct fid *fid, struct fid *bfid, uint64_t flags)
+DIRECT_FN
+STATIC int psmx2_sep_bind(struct fid *fid, struct fid *bfid, uint64_t flags)
 {
 	struct psmx2_fid_sep *sep;
 	int i, err = 0;
@@ -833,29 +861,29 @@ static int psmx2_sep_bind(struct fid *fid, struct fid *bfid, uint64_t flags)
 	return err;
 }
 
-static int psmx2_tx_context(struct fid_ep *ep, int index, struct fi_tx_attr *attr,
+DIRECT_FN
+STATIC int psmx2_tx_context(struct fid_ep *ep, int index, struct fi_tx_attr *attr,
 			    struct fid_ep **tx_ep, void *context)
 {
 	struct psmx2_fid_sep *sep;
 
 	sep = container_of(ep, struct psmx2_fid_sep, ep);
 
-	if (index < 0 || index > sep->ctxt_cnt)
-		return -FI_EINVAL;
+	assert(index >= 0 && index < sep->ctxt_cnt);
 
 	*tx_ep = &sep->ctxts[index].ep->ep;
 	return 0;
 }
 
-static int psmx2_rx_context(struct fid_ep *ep, int index, struct fi_rx_attr *attr,
+DIRECT_FN
+STATIC int psmx2_rx_context(struct fid_ep *ep, int index, struct fi_rx_attr *attr,
 			    struct fid_ep **rx_ep, void *context)
 {
 	struct psmx2_fid_sep *sep;
 
 	sep = container_of(ep, struct psmx2_fid_sep, ep);
 
-	if (index < 0 || index > sep->ctxt_cnt)
-		return -FI_EINVAL;
+	assert(index >= 0 && index < sep->ctxt_cnt);
 
 	*rx_ep = &sep->ctxts[index].ep->ep;
 	return 0;
@@ -900,6 +928,7 @@ static struct fi_ops_ep psmx2_sep_ops = {
 	.tx_size_left = fi_no_tx_size_left,
 };
 
+DIRECT_FN
 int psmx2_sep_open(struct fid_domain *domain, struct fi_info *info,
 		   struct fid_ep **sep, void *context)
 {
@@ -920,18 +949,18 @@ int psmx2_sep_open(struct fid_domain *domain, struct fi_info *info,
 		goto errout;
 
 	if (info && info->ep_attr) {
-		if (info->ep_attr->tx_ctx_cnt > psmx2_env.sep_trx_ctxt) {
+		if (info->ep_attr->tx_ctx_cnt > psmx2_hfi_info.max_trx_ctxt) {
 			FI_WARN(&psmx2_prov, FI_LOG_EP_CTRL,
 				"tx_ctx_cnt %"PRIu64" exceed limit %d.\n",
 				info->ep_attr->tx_ctx_cnt,
-				psmx2_env.sep_trx_ctxt);
+				psmx2_hfi_info.max_trx_ctxt);
 			goto errout;
 		}
-		if (info->ep_attr->rx_ctx_cnt > psmx2_env.sep_trx_ctxt) {
+		if (info->ep_attr->rx_ctx_cnt > psmx2_hfi_info.max_trx_ctxt) {
 			FI_WARN(&psmx2_prov, FI_LOG_EP_CTRL,
 				"rx_ctx_cnt %"PRIu64" exceed limit %d.\n",
 				info->ep_attr->rx_ctx_cnt,
-				psmx2_env.sep_trx_ctxt);
+				psmx2_hfi_info.max_trx_ctxt);
 			goto errout;
 		}
 		ctxt_cnt = info->ep_attr->tx_ctx_cnt;
@@ -969,7 +998,9 @@ int psmx2_sep_open(struct fid_domain *domain, struct fi_info *info,
 	}
 
 	for (i = 0; i < ctxt_cnt; i++) {
-		trx_ctxt = psmx2_trx_ctxt_alloc(domain_priv, src_addr, i);
+		trx_ctxt = psmx2_trx_ctxt_alloc(domain_priv, src_addr,
+						(ctxt_cnt > 1) ? i : -1,
+						PSMX2_TX_RX);
 		if (!trx_ctxt) {
 			err = -FI_ENOMEM;
 			goto errout_free_ctxt;
@@ -978,14 +1009,13 @@ int psmx2_sep_open(struct fid_domain *domain, struct fi_info *info,
 		sep_priv->ctxts[i].trx_ctxt = trx_ctxt;
 
 		err = psmx2_ep_open_internal(domain_priv, info, &ep_priv, context,
-					     trx_ctxt, 0);
+					     trx_ctxt, PSMX2_TX_RX);
 		if (err)
 			goto errout_free_ctxt;
 
 		/* override the ops so the fid can't be closed individually */
 		ep_priv->ep.fid.ops = &psmx2_fi_ops_sep_ctxt;
 
-		trx_ctxt->ep = ep_priv;
 		sep_priv->ctxts[i].ep = ep_priv;
 	}
 
@@ -1002,19 +1032,14 @@ int psmx2_sep_open(struct fid_domain *domain, struct fi_info *info,
 				   ((uintptr_t)sep_priv & 0xFFFF);
 
 	sep_priv->id = ofi_atomic_inc32(&domain_priv->sep_cnt);
+	for (i = 0; i < ctxt_cnt; i++)
+		sep_priv->ctxts[i].ep->sep_id = sep_priv->id;
 
-	psmx2_lock(&domain_priv->sep_lock, 1);
+	domain_priv->sep_lock_fn(&domain_priv->sep_lock, 1);
 	dlist_insert_before(&sep_priv->entry, &domain_priv->sep_list);
-	psmx2_unlock(&domain_priv->sep_lock, 1);
+	domain_priv->sep_unlock_fn(&domain_priv->sep_lock, 1);
 
-	psmx2_lock(&domain_priv->trx_ctxt_lock, 1);
-	for (i = 0; i< ctxt_cnt; i++) {
-		dlist_insert_before(&sep_priv->ctxts[i].trx_ctxt->entry,
-				    &domain_priv->trx_ctxt_list);
-	}
-	psmx2_unlock(&domain_priv->trx_ctxt_lock, 1);
-
-	ep_name.epid = domain_priv->base_trx_ctxt->psm2_epid;
+	ep_name.epid = sep_priv->ctxts[0].trx_ctxt->psm2_epid;
 	ep_name.sep_id = sep_priv->id;
 	ep_name.type = sep_priv->type;
 
@@ -1023,16 +1048,21 @@ int psmx2_sep_open(struct fid_domain *domain, struct fi_info *info,
 
 	psmx2_domain_acquire(domain_priv);
 	*sep = &sep_priv->ep;
+
+	/* Make sure the AM handler is installed to answer SEP query */
+	psmx2_am_init(sep_priv->ctxts[0].trx_ctxt);
+
 	return 0;
 
 errout_free_ctxt:
 	while (i) {
+		if (sep_priv->ctxts[i].trx_ctxt)
+			psmx2_trx_ctxt_free(sep_priv->ctxts[i].trx_ctxt,
+					    PSMX2_TX_RX);
+
 		if (sep_priv->ctxts[i].ep)
 			psmx2_ep_close_internal(sep_priv->ctxts[i].ep);
 
-		if (sep_priv->ctxts[i].trx_ctxt)
-			psmx2_trx_ctxt_free(sep_priv->ctxts[i].trx_ctxt);
-
 		i--;
 	}
 
