@@ -1,6 +1,7 @@
 /*
  * Copyright (c) 2014 Intel Corporation, Inc.  All rights reserved.
  * Copyright (c) 2016 Cisco Systems, Inc. All rights reserved.
+ * Copyright (c) 2017 DataDirect Networks, Inc. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -48,20 +49,23 @@
 #include <netdb.h>
 #include <netinet/in.h>
 
-#include <fi.h>
+#include <ofi.h>
+#include <ofi_atom.h>
 #include <ofi_atomic.h>
-#include <fi_enosys.h>
-#include <fi_indexer.h>
-#include <fi_rbuf.h>
-#include <fi_list.h>
-#include <fi_file.h>
-#include <fi_osd.h>
-#include "fi_util.h"
+#include <ofi_mr.h>
+#include <ofi_enosys.h>
+#include <ofi_indexer.h>
+#include <ofi_rbuf.h>
+#include <ofi_list.h>
+#include <ofi_file.h>
+#include <ofi_osd.h>
+#include <ofi_util.h>
 
 #ifndef _SOCK_H_
 #define _SOCK_H_
 
-#define SOCK_EP_MAX_MSG_SZ (1<<23)
+/* 4k allocated for all sockets headers */
+#define SOCK_EP_MAX_MSG_SZ (OFI_MAX_SOCKET_BUF_SIZE - 4096)
 #define SOCK_EP_MAX_INJECT_SZ ((1<<8) - 1)
 #define SOCK_EP_MAX_BUFF_RECV (1<<26)
 #define SOCK_EP_MAX_ORDER_RAW_SZ SOCK_EP_MAX_MSG_SZ
@@ -90,6 +94,7 @@
 #define SOCK_CQ_DEF_SZ (1<<8)
 #define SOCK_AV_DEF_SZ (1<<8)
 #define SOCK_CMAP_DEF_SZ (1<<10)
+#define SOCK_EPOLL_WAIT_EVENTS 32
 
 #define SOCK_CQ_DATA_SIZE (sizeof(uint64_t))
 #define SOCK_TAG_SIZE (sizeof(uint64_t))
@@ -100,41 +105,15 @@
 #define SOCK_EP_MAX_RETRY (5)
 #define SOCK_EP_MAX_CM_DATA_SZ (256)
 #define SOCK_CM_DEF_BACKLOG (128)
+#define SOCK_CM_DEF_TIMEOUT (15000)
 #define SOCK_CM_DEF_RETRY (5)
 #define SOCK_CM_CONN_IN_PROGRESS ((struct sock_conn *)(0x1L))
 
-#define SOCK_EP_RDM_PRI_CAP (FI_MSG | FI_RMA | FI_TAGGED | FI_ATOMICS |	\
-			 FI_NAMED_RX_CTX | \
-			 FI_DIRECTED_RECV | \
-			 FI_READ | FI_WRITE | FI_RECV | FI_SEND | \
-			 FI_REMOTE_READ | FI_REMOTE_WRITE)
-
-#define SOCK_EP_RDM_SEC_CAP (FI_MULTI_RECV | FI_SOURCE | FI_RMA_EVENT | \
-			 FI_SHARED_AV | FI_FENCE | FI_TRIGGER)
-
-#define SOCK_EP_RDM_CAP (SOCK_EP_RDM_PRI_CAP | SOCK_EP_RDM_SEC_CAP)
-
-#define SOCK_EP_MSG_PRI_CAP SOCK_EP_RDM_PRI_CAP
-
-#define SOCK_EP_MSG_SEC_CAP SOCK_EP_RDM_SEC_CAP
-
-#define SOCK_EP_MSG_CAP (SOCK_EP_MSG_PRI_CAP | SOCK_EP_MSG_SEC_CAP)
-
-#define SOCK_EP_DGRAM_PRI_CAP (FI_MSG | FI_TAGGED | \
-			   FI_NAMED_RX_CTX | FI_DIRECTED_RECV | \
-			   FI_RECV | FI_SEND)
-
-#define SOCK_EP_DGRAM_SEC_CAP (FI_MULTI_RECV | FI_SOURCE | FI_SHARED_AV | \
-			   FI_FENCE | FI_TRIGGER)
-
-#define SOCK_EP_DGRAM_CAP (SOCK_EP_DGRAM_PRI_CAP | SOCK_EP_DGRAM_SEC_CAP)
-
-#define SOCK_EP_MSG_ORDER (FI_ORDER_RAR | FI_ORDER_RAW | FI_ORDER_RAS|	\
-			   FI_ORDER_WAR | FI_ORDER_WAW | FI_ORDER_WAS |	\
+#define SOCK_EP_MSG_ORDER (OFI_ORDER_RAR_SET | OFI_ORDER_RAW_SET | FI_ORDER_RAS| \
+			   OFI_ORDER_WAR_SET | OFI_ORDER_WAW_SET | FI_ORDER_WAS | \
 			   FI_ORDER_SAR | FI_ORDER_SAW | FI_ORDER_SAS)
 
 #define SOCK_EP_COMP_ORDER (FI_ORDER_STRICT | FI_ORDER_DATA)
-#define SOCK_EP_DEFAULT_OP_FLAGS (FI_TRANSMIT_COMPLETE)
 
 #define SOCK_EP_CQ_FLAGS (FI_SEND | FI_TRANSMIT | FI_RECV | \
 			FI_SELECTIVE_COMPLETION)
@@ -153,7 +132,7 @@
 #define SOCK_PE_COMM_BUFF_SZ (1024)
 #define SOCK_PE_OVERFLOW_COMM_BUFF_SZ (128)
 
-/* it must be adjusted if error data size in CQ/EQ 
+/* it must be adjusted if error data size in CQ/EQ
  * will be larger than SOCK_EP_MAX_CM_DATA_SZ */
 #define SOCK_MAX_ERR_CQ_EQ_DATA_SZ SOCK_EP_MAX_CM_DATA_SZ
 
@@ -162,31 +141,36 @@ enum {
 	SOCK_SIGNAL_WR_FD
 };
 
-#define SOCK_MAJOR_VERSION 2
-#define SOCK_MINOR_VERSION 0
+enum {
+	SOCK_OPTS_NONBLOCK  = 1<<0,
+	SOCK_OPTS_KEEPALIVE = 1<<1
+};
 
-#define SOCK_WIRE_PROTO_VERSION (1)
+#define SOCK_WIRE_PROTO_VERSION (2)
+
+extern struct fi_info sock_dgram_info;
+extern struct fi_info sock_msg_info;
+
+extern struct util_prov sock_util_prov;
+extern struct fi_domain_attr sock_domain_attr;
+extern struct fi_fabric_attr sock_fabric_attr;
+extern struct fi_tx_attr sock_msg_tx_attr;
+extern struct fi_tx_attr sock_rdm_tx_attr;
+extern struct fi_tx_attr sock_dgram_tx_attr;
+extern struct fi_rx_attr sock_msg_rx_attr;
+extern struct fi_rx_attr sock_rdm_rx_attr;
+extern struct fi_rx_attr sock_dgram_rx_attr;
+extern struct fi_ep_attr sock_msg_ep_attr;
+extern struct fi_ep_attr sock_rdm_ep_attr;
+extern struct fi_ep_attr sock_dgram_ep_attr;
+extern struct fi_tx_attr sock_stx_attr;
+extern struct fi_rx_attr sock_srx_attr;
 
 struct sock_service_entry {
 	int service;
 	struct dlist_entry entry;
 };
 
-#ifdef HAVE_EPOLL
-struct sock_epoll_set {
-	int fd;
-	int size;
-	int used;
-	struct epoll_event *events;
-};
-#else
-struct sock_epoll_set {
-	struct pollfd *pollfds;
-	int size;
-	int used;
-};
-#endif
-
 struct sock_fabric {
 	struct fid_fabric fab_fid;
 	ofi_atomic32_t ref;
@@ -202,7 +186,7 @@ struct sock_conn {
 	int sock_fd;
 	int connected;
 	int address_published;
-	struct sockaddr_in addr;
+	union ofi_sock_ip addr;
 	struct sock_pe_entry *rx_pe_entry;
 	struct sock_pe_entry *tx_pe_entry;
 	struct sock_ep_attr *ep_attr;
@@ -212,12 +196,33 @@ struct sock_conn {
 
 struct sock_conn_map {
 	struct sock_conn *table;
-	struct sock_epoll_set epoll_set;
+	ofi_epoll_t epoll_set;
+	void **epoll_ctxs;
+	int epoll_ctxs_sz;
 	int used;
 	int size;
 	fastlock_t lock;
 };
 
+struct sock_conn_listener {
+	ofi_epoll_t epollfd;
+	struct fd_signal signal;
+	fastlock_t signal_lock; /* acquire before map lock */
+	pthread_t listener_thread;
+	int do_listen;
+	bool removed_from_epollfd;
+};
+
+struct sock_ep_cm_head {
+	ofi_epoll_t epollfd;
+	struct fd_signal signal;
+	pthread_mutex_t signal_lock;
+	pthread_t listener_thread;
+	struct dlist_entry msg_list;
+	int do_listen;
+	bool removed_from_epollfd;
+};
+
 struct sock_domain {
 	struct fi_info		info;
 	struct fid_domain	dom_fid;
@@ -233,6 +238,8 @@ struct sock_domain {
 	struct sock_pe		*pe;
 	struct dlist_entry	dom_list_entry;
 	struct fi_domain_attr	attr;
+	struct sock_conn_listener conn_listener;
+	struct sock_ep_cm_head cm_head;
 };
 
 /* move to fi_trigger.h when removing experimental tag from work queues */
@@ -330,7 +337,7 @@ struct sock_mr {
 };
 
 struct sock_av_addr {
-	struct sockaddr_storage addr;
+	union ofi_sock_ip addr;
 	uint8_t valid;
 	uint8_t reserved[7];
 };
@@ -356,6 +363,7 @@ struct sock_av {
 	int    shared;
 	struct dlist_entry ep_list;
 	fastlock_t list_lock;
+	fastlock_t table_lock;
 };
 
 struct sock_fid_list {
@@ -434,8 +442,6 @@ struct sock_op_send {
 	uint64_t context;
 	uint64_t dest_addr;
 	uint64_t buf;
-	uint64_t tag;
-	uint64_t data;
 	struct sock_ep *ep;
 	struct sock_conn *conn;
 };
@@ -504,24 +510,28 @@ struct sock_comp {
 	struct sock_eq *eq;
 };
 
-struct sock_cm_entry {
-	int sock;
-	int do_listen;
-	int signal_fds[2];
-	uint64_t next_msg_id;
-	fastlock_t lock;
-	int is_connected;
-	pthread_t listener_thread;
-	struct dlist_entry msg_list;
+enum sock_cm_state {
+	SOCK_CM_STATE_DISCONNECTED = 0,
+	SOCK_CM_STATE_REQUESTED,
+	SOCK_CM_STATE_CONNECTED,
 };
 
-struct sock_conn_listener {
+struct sock_pep_cm_entry {
 	int sock;
 	int do_listen;
-	int is_ready;
 	int signal_fds[2];
 	pthread_t listener_thread;
-	char service[NI_MAXSERV];
+};
+
+struct sock_ep_cm_entry {
+	int sock;
+	fastlock_t lock;
+	enum sock_cm_state state;
+};
+
+struct sock_conn_handle {
+	int sock;
+	int do_listen;
 };
 
 struct sock_ep_attr {
@@ -552,19 +562,18 @@ struct sock_ep_attr {
 	struct fi_ep_attr ep_attr;
 
 	enum fi_ep_type ep_type;
-	struct sockaddr_in *src_addr;
-	struct sockaddr_in *dest_addr;
+	union ofi_sock_ip *src_addr;
+	union ofi_sock_ip *dest_addr;
 	uint16_t msg_src_port;
 	uint16_t msg_dest_port;
 
 	uint64_t peer_fid;
 	uint16_t key;
 	int is_enabled;
-	struct sock_cm_entry cm;
-	struct sock_conn_listener listener;
+	struct sock_ep_cm_entry cm;
+	struct sock_conn_handle conn_handle;
 	fastlock_t lock;
 
-	struct index_map conn_idm;
 	struct index_map av_idm;
 	struct sock_conn_map cmap;
 };
@@ -581,8 +590,9 @@ struct sock_pep {
 	struct fid_pep	pep;
 	struct sock_fabric *sock_fab;
 
-	struct sock_cm_entry cm;
-	struct sockaddr_in src_addr;
+	struct sock_ep_cm_head cm_head;
+	struct sock_pep_cm_entry cm;
+	union ofi_sock_ip src_addr;
 	struct fi_info info;
 	struct sock_eq *eq;
 	int name_set;
@@ -646,6 +656,8 @@ struct sock_rx_ctx {
 	struct dlist_entry ep_list;
 	fastlock_t lock;
 
+	struct dlist_entry *progress_start;
+
 	struct fi_rx_attr attr;
 	struct sock_rx_entry *rx_entry_pool;
 	struct slist pool_list;
@@ -813,6 +825,7 @@ struct sock_pe_entry {
 	uint8_t is_error;
 	uint8_t mr_checked;
 	uint8_t is_pool_entry;
+	uint8_t completion_reported;
 	uint8_t reserved[3];
 
 	uint64_t done_len;
@@ -841,8 +854,8 @@ struct sock_pe {
 	int signal_fds[2];
 	uint64_t waittime;
 
-	struct util_buf_pool *pe_rx_pool;
-	struct util_buf_pool *atomic_rx_pool;
+	struct ofi_bufpool *pe_rx_pool;
+	struct ofi_bufpool *atomic_rx_pool;
 	struct dlist_entry free_list;
 	struct dlist_entry busy_list;
 	struct dlist_entry pool_list;
@@ -853,7 +866,7 @@ struct sock_pe {
 	pthread_t progress_thread;
 	volatile int do_progress;
 	struct sock_pe_entry *pe_atomic;
-	struct sock_epoll_set epoll_set;
+	ofi_epoll_t epoll_set;
 };
 
 typedef int (*sock_cq_report_fn) (struct sock_cq *cq, fi_addr_t addr,
@@ -877,11 +890,12 @@ struct sock_cq {
 	struct ofi_ringbuffd cq_rbfd;
 	struct ofi_ringbuf cqerr_rb;
 	struct dlist_entry overflow_list;
-	fastlock_t lock;
-	fastlock_t list_lock;
+	pthread_mutex_t lock;
+	pthread_mutex_t list_lock;
 
 	struct fid_wait *waitset;
 	int signal;
+	ofi_atomic32_t signaled;
 
 	struct dlist_entry ep_list;
 	struct dlist_entry rx_list;
@@ -900,7 +914,7 @@ struct sock_conn_hdr {
 
 struct sock_conn_req {
 	struct sock_conn_hdr hdr;
-	struct sockaddr_in src_addr;
+	union ofi_sock_ip src_addr;
 	uint64_t caps;
 	char cm_data[0];
 };
@@ -912,25 +926,31 @@ enum {
 	SOCK_CONN_SHUTDOWN,
 };
 
+enum sock_conn_handle_state {
+	SOCK_CONN_HANDLE_ACTIVE,
+	SOCK_CONN_HANDLE_ACCEPTED,
+	SOCK_CONN_HANDLE_REJECTED,
+	SOCK_CONN_HANDLE_DELETED,
+	SOCK_CONN_HANDLE_FINALIZING,
+	SOCK_CONN_HANDLE_FINALIZED,
+};
+
 struct sock_conn_req_handle {
 	struct fid handle;
 	struct sock_conn_req *req;
 	int sock_fd;
-	int is_accepted;
+	uint8_t monitored;
+	enum sock_conn_handle_state state;
+	pthread_mutex_t	finalized_mutex;
+	pthread_cond_t	finalized_cond;
 	struct sock_pep *pep;
 	struct sock_ep *ep;
 	size_t paramlen;
-	pthread_t req_handler;
-	struct sockaddr_in dest_addr;
+	union ofi_sock_ip dest_addr;
 	struct dlist_entry entry;
 	char cm_data[SOCK_EP_MAX_CM_DATA_SZ];
 };
 
-struct sock_host_list_entry {
-	char hostname[HOST_NAME_MAX];
-	struct slist_entry entry;
-};
-
 union sock_tx_op {
 	struct sock_msg {
 		struct sock_op_send op;
@@ -968,34 +988,20 @@ union sock_tx_op {
 };
 #define SOCK_EP_TX_ENTRY_SZ (sizeof(union sock_tx_op))
 
-int sock_verify_info(uint32_t version, struct fi_info *hints);
-int sock_verify_fabric_attr(struct fi_fabric_attr *attr);
-int sock_verify_domain_attr(uint32_t version, struct fi_domain_attr *attr);
-
 size_t sock_get_tx_size(size_t size);
-int sock_rdm_verify_ep_attr(struct fi_ep_attr *ep_attr, struct fi_tx_attr *tx_attr,
-			    struct fi_rx_attr *rx_attr);
-int sock_dgram_verify_ep_attr(struct fi_ep_attr *ep_attr, struct fi_tx_attr *tx_attr,
-			      struct fi_rx_attr *rx_attr);
-int sock_msg_verify_ep_attr(struct fi_ep_attr *ep_attr, struct fi_tx_attr *tx_attr,
-			    struct fi_rx_attr *rx_attr);
-int sock_get_src_addr(struct sockaddr_in *dest_addr,
-		      struct sockaddr_in *src_addr);
-int sock_get_src_addr_from_hostname(struct sockaddr_in *src_addr, const char *service);
+int sock_get_src_addr(union ofi_sock_ip *dest_addr,
+		      union ofi_sock_ip *src_addr);
+int sock_get_src_addr_from_hostname(union ofi_sock_ip *src_addr,
+				    const char *service, uint16_t sa_family);
 
 struct fi_info *sock_fi_info(uint32_t version, enum fi_ep_type ep_type,
-			     struct fi_info *hints, void *src_addr,
+			     const struct fi_info *hints, void *src_addr,
 			     void *dest_addr);
-int sock_msg_fi_info(uint32_t version, void *src_addr, void *dest_addr,
-		     struct fi_info *hints, struct fi_info **info);
-int sock_dgram_fi_info(uint32_t version, void *src_addr, void *dest_addr,
-		       struct fi_info *hints, struct fi_info **info);
-int sock_rdm_fi_info(uint32_t version, void *src_addr, void *dest_addr,
-		     struct fi_info *hints, struct fi_info **info);
 void free_fi_info(struct fi_info *info);
 
 int sock_msg_getinfo(uint32_t version, const char *node, const char *service,
-		uint64_t flags, struct fi_info *hints, struct fi_info **info);
+		uint64_t flags, const struct fi_info *hints,
+		struct fi_info **info);
 
 int sock_domain(struct fid_fabric *fabric, struct fi_info *info,
 		struct fid_domain **dom, void *context);
@@ -1113,24 +1119,25 @@ int sock_wait_close(fid_t fid);
 int sock_av_open(struct fid_domain *domain, struct fi_av_attr *attr,
 		 struct fid_av **av, void *context);
 int sock_av_compare_addr(struct sock_av *av, fi_addr_t addr1, fi_addr_t addr2);
-int sock_av_get_addr_index(struct sock_av *av, struct sockaddr_in *addr);
+int sock_av_get_addr_index(struct sock_av *av, union ofi_sock_ip *addr);
 
 struct sock_conn *sock_ep_lookup_conn(struct sock_ep_attr *attr, fi_addr_t index,
-                                      struct sockaddr_in *addr);
+				      union ofi_sock_ip *addr);
 int sock_ep_get_conn(struct sock_ep_attr *ep_attr, struct sock_tx_ctx *tx_ctx,
 		     fi_addr_t index, struct sock_conn **pconn);
 void sock_ep_remove_conn(struct sock_ep_attr *ep_attr, struct sock_conn *conn);
-struct sock_conn *sock_ep_connect(struct sock_ep_attr *attr, fi_addr_t index);
+int sock_ep_connect(struct sock_ep_attr *attr, fi_addr_t index,
+		    struct sock_conn **conn);
 ssize_t sock_conn_send_src_addr(struct sock_ep_attr *ep_attr, struct sock_tx_ctx *tx_ctx,
 				struct sock_conn *conn);
 int sock_conn_listen(struct sock_ep_attr *ep_attr);
+int sock_conn_start_listener_thread(struct sock_conn_listener *conn_listener);
+int sock_conn_stop_listener_thread(struct sock_conn_listener *conn_listener);
 void sock_conn_map_destroy(struct sock_ep_attr *ep_attr);
 void sock_conn_release_entry(struct sock_conn_map *map, struct sock_conn *conn);
-void sock_set_sockopts(int sock);
+void sock_set_sockopts(int sock, int sock_opts);
 int fd_set_nonblock(int fd);
-void sock_set_sockopt_reuseaddr(int sock);
 int sock_conn_map_init(struct sock_ep *ep, int init_size);
-void sock_set_sockopts_conn(int sock);
 
 struct sock_pe *sock_pe_init(struct sock_domain *domain);
 void sock_pe_add_tx_ctx(struct sock_pe *pe, struct sock_tx_ctx *ctx);
@@ -1138,6 +1145,9 @@ void sock_pe_add_rx_ctx(struct sock_pe *pe, struct sock_rx_ctx *ctx);
 void sock_pe_signal(struct sock_pe *pe);
 void sock_pe_poll_add(struct sock_pe *pe, int fd);
 void sock_pe_poll_del(struct sock_pe *pe, int fd);
+
+int sock_pe_progress_ep_rx(struct sock_pe *pe, struct sock_ep_attr *ep_attr);
+int sock_pe_progress_ep_tx(struct sock_pe *pe, struct sock_ep_attr *ep_attr);
 int sock_pe_progress_rx_ctx(struct sock_pe *pe, struct sock_rx_ctx *rx_ctx);
 int sock_pe_progress_tx_ctx(struct sock_pe *pe, struct sock_tx_ctx *tx_ctx);
 void sock_pe_remove_tx_ctx(struct sock_tx_ctx *tx_ctx);
@@ -1203,16 +1213,15 @@ ssize_t sock_queue_msg_op(struct fid_ep *ep, const struct fi_msg *msg,
 ssize_t sock_queue_cntr_op(struct fi_deferred_work *work, uint64_t flags);
 void sock_cntr_check_trigger_list(struct sock_cntr *cntr);
 
-int sock_epoll_create(struct sock_epoll_set *set, int size);
-int sock_epoll_add(struct sock_epoll_set *set, int fd);
-int sock_epoll_del(struct sock_epoll_set *set, int fd);
-int sock_epoll_wait(struct sock_epoll_set *set, int timeout);
-int sock_epoll_get_fd_at_index(struct sock_epoll_set *set, int index);
-void sock_epoll_close(struct sock_epoll_set *set);
-
 static inline size_t sock_rx_avail_len(struct sock_rx_entry *rx_entry)
 {
 	return rx_entry->total_len - rx_entry->used;
 }
 
+int sock_ep_cm_start_thread(struct sock_ep_cm_head *cm_head);
+void sock_ep_cm_signal(struct sock_ep_cm_head *cm_head);
+void sock_ep_cm_stop_thread(struct sock_ep_cm_head *cm_head);
+void sock_ep_cm_wait_handle_finalized(struct sock_ep_cm_head *cm_head,
+                                      struct sock_conn_req_handle *handle);
+
 #endif
