@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2017, Cisco Systems, Inc. All rights reserved.
+ * Copyright (c) 2014-2019, Cisco Systems, Inc. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -40,6 +40,7 @@
 #include <pthread.h>
 
 #include <rdma/providers/fi_log.h>
+#include <ofi_epoll.h>
 
 #include "usdf_progress.h"
 #include "usd.h"
@@ -120,7 +121,7 @@ struct usdf_fabric {
 	/* progression */
 	pthread_t fab_thread;
 	int fab_exit;
-	int fab_epollfd;
+	ofi_epoll_t fab_epollfd;
 	int fab_eventfd;
 	struct usdf_poll_item fab_poll_item;
 
@@ -147,11 +148,6 @@ struct usdf_domain {
 	TAILQ_HEAD(,usdf_tx) dom_tx_ready;
 	TAILQ_HEAD(,usdf_cq_hard) dom_hcq_list;
 
-	struct usdf_rdm_connection **dom_rdc_hashtab;
-	SLIST_HEAD(,usdf_rdm_connection) dom_rdc_free;
-	ofi_atomic32_t dom_rdc_free_cnt;
-	size_t dom_rdc_total;
-
 	/* used only by connected endpoints */
 	struct usdf_ep **dom_peer_tab;
 	uint32_t dom_next_peer;
@@ -496,7 +492,7 @@ int usdf_query_atomic(struct fid_domain *domain, enum fi_datatype datatype,
 int usdf_domain_getname(uint32_t version, struct usd_device_attrs *dap,
 			char **name);
 bool usdf_domain_checkname(uint32_t version, struct usd_device_attrs *dap,
-			char *hint);
+			const char *hint);
 
 /* fi_ops_mr */
 int usdf_reg_mr(struct fid *fid, const void *buf, size_t len,
@@ -516,14 +512,12 @@ void usdf_setup_fake_ibv_provider(void);
 int usdf_pep_steal_socket(struct usdf_pep *pep, int *is_bound, int *sock_o);
 
 /* Utility functions */
-int usdf_catch_dom_attr(uint32_t version, struct fi_info *hints,
+int usdf_catch_dom_attr(uint32_t version, const struct fi_info *hints,
 			struct fi_domain_attr *dom_attr);
-int usdf_catch_tx_attr(uint32_t version, struct fi_tx_attr *tx_attr);
-int usdf_catch_rx_attr(uint32_t version, struct fi_rx_attr *rx_attr);
-int usdf_check_mr_mode(uint32_t version, struct fi_info *hints,
-		       uint64_t prov_mode);
-struct sockaddr_in *usdf_format_to_sin(struct fi_info *info, const void *addr);
-void *usdf_sin_to_format(struct fi_info *info, void *addr, size_t *len);
-void usdf_free_sin_if_needed(struct fi_info *info, struct sockaddr_in *sin);
+int usdf_catch_tx_attr(uint32_t version, const struct fi_tx_attr *tx_attr);
+int usdf_catch_rx_attr(uint32_t version, const struct fi_rx_attr *rx_attr);
+struct sockaddr_in *usdf_format_to_sin(const struct fi_info *info, const void *addr);
+void *usdf_sin_to_format(const struct fi_info *info, void *addr, size_t *len);
+void usdf_free_sin_if_needed(const struct fi_info *info, struct sockaddr_in *sin);
 
 #endif /* _USDF_H_ */
