@@ -33,8 +33,8 @@
 #include <stdlib.h>
 #include <string.h>
 
-#include <fi_enosys.h>
-#include <fi_util.h>
+#include <ofi_enosys.h>
+#include <ofi_util.h>
 
 
 int ofi_domain_bind_eq(struct util_domain *domain, struct util_eq *eq)
@@ -55,6 +55,9 @@ int ofi_domain_close(struct util_domain *domain)
 	if (ofi_atomic_get32(&domain->ref))
 		return -FI_EBUSY;
 
+	if (domain->mr_map.rbtree)
+		ofi_mr_map_close(&domain->mr_map);
+
 	fastlock_acquire(&domain->fabric->lock);
 	dlist_remove(&domain->list_entry);
 	fastlock_release(&domain->fabric->lock);
@@ -83,6 +86,8 @@ static int util_domain_init(struct util_domain *domain,
 	domain->addr_format = info->addr_format;
 	domain->av_type = info->domain_attr->av_type;
 	domain->name = strdup(info->domain_attr->name);
+	domain->threading = info->domain_attr->threading;
+	domain->data_progress = info->domain_attr->data_progress;
 	return domain->name ? 0 : -FI_ENOMEM;
 }
 
@@ -96,10 +101,8 @@ int ofi_domain_init(struct fid_fabric *fabric_fid, const struct fi_info *info,
 	domain->fabric = fabric;
 	domain->prov = fabric->prov;
 	ret = util_domain_init(domain, info);
-	if (ret) {
-		free(domain);
+	if (ret)
 		return ret;
-	}
 
 	domain->domain_fid.fid.fclass = FI_CLASS_DOMAIN;
 	domain->domain_fid.fid.context = context;
@@ -108,6 +111,13 @@ int ofi_domain_init(struct fid_fabric *fabric_fid, const struct fi_info *info,
 	 */
 	domain->domain_fid.mr = &util_domain_mr_ops;
 
+	ret = ofi_mr_map_init(domain->prov, info->domain_attr->mr_mode,
+			      &domain->mr_map);
+	if (ret) {
+		(void) ofi_domain_close(domain);
+		return ret;
+	}
+
 	fastlock_acquire(&fabric->lock);
 	dlist_insert_tail(&domain->list_entry, &fabric->domain_list);
 	fastlock_release(&fabric->lock);
