@@ -33,15 +33,47 @@
 #include <stdlib.h>
 #include <string.h>
 
-#include <fi_enosys.h>
-#include <fi_util.h>
+#include <ofi_enosys.h>
+#include <ofi_util.h>
 
+void ofi_eq_handle_err_entry(uint32_t api_version, uint64_t flags,
+			     struct fi_eq_err_entry *err_entry,
+			     struct fi_eq_err_entry *user_err_entry)
+{
+	if ((FI_VERSION_GE(api_version, FI_VERSION(1, 5)))
+	    && user_err_entry->err_data && user_err_entry->err_data_size
+	    && err_entry->err_data && err_entry->err_data_size) {
+		void *err_data = user_err_entry->err_data;
+		size_t err_data_size = MIN(err_entry->err_data_size,
+					   user_err_entry->err_data_size);
 
-static ssize_t util_eq_read(struct fid_eq *eq_fid, uint32_t *event,
-			    void *buf, size_t len, uint64_t flags)
+		memcpy(err_data, err_entry->err_data, err_data_size);
+
+		*user_err_entry = *err_entry;
+		user_err_entry->err_data = err_data;
+		user_err_entry->err_data_size = err_data_size;
+
+		if (!(flags & FI_PEEK)) {
+			free(err_entry->err_data);
+			err_entry->err_data = NULL;
+			err_entry->err_data_size = 0;
+		}
+	} else {
+		*user_err_entry = *err_entry;
+	}
+
+	if (!(flags & FI_PEEK)) {
+		err_entry->err = 0;
+		err_entry->prov_errno = 0;
+	}
+}
+
+ssize_t ofi_eq_read(struct fid_eq *eq_fid, uint32_t *event,
+		    void *buf, size_t len, uint64_t flags)
 {
 	struct util_eq *eq;
 	struct util_event *entry;
+	struct fi_eq_err_entry *err_entry;
 	ssize_t ret;
 
 	eq = container_of(eq_fid, struct util_eq, eq_fid);
@@ -64,8 +96,23 @@ static ssize_t util_eq_read(struct fid_eq *eq_fid, uint32_t *event,
 	if (event)
 		*event = entry->event;
 	if (buf) {
-		ret = MIN(len, (size_t)entry->size);
-		memcpy(buf, entry->data, ret);
+		if (flags & UTIL_FLAG_ERROR) {
+			free(eq->saved_err_data);
+			eq->saved_err_data = NULL;
+
+			assert((size_t) entry->size == sizeof(*err_entry));
+			err_entry = (struct fi_eq_err_entry *) entry->data;
+
+			ofi_eq_handle_err_entry(eq->fabric->fabric_fid.api_version,
+						flags, err_entry, buf);
+			ret = entry->size;
+
+			if (!(flags & FI_PEEK))
+				eq->saved_err_data = err_entry->err_data;
+		} else {
+			ret = MIN(len, (size_t)entry->size);
+			memcpy(buf, entry->data, ret);
+		}
 	}  else {
 		ret = 0;
 	}
@@ -79,26 +126,25 @@ out:
 	return ret;
 }
 
-static ssize_t util_eq_readerr(struct fid_eq *eq_fid, struct fi_eq_err_entry *buf,
-			       uint64_t flags)
+ssize_t ofi_eq_readerr(struct fid_eq *eq_fid, struct fi_eq_err_entry *buf,
+		       uint64_t flags)
 {
-
-	return util_eq_read(eq_fid, NULL, buf, sizeof(*buf),
-			    flags | UTIL_FLAG_ERROR);
+	return fi_eq_read(eq_fid, NULL, buf, sizeof(*buf),
+			  flags | UTIL_FLAG_ERROR);
 }
 
-static ssize_t util_eq_write(struct fid_eq *eq_fid, uint32_t event,
-			     const void *buf, size_t len, uint64_t flags)
+ssize_t ofi_eq_write(struct fid_eq *eq_fid, uint32_t event,
+		     const void *buf, size_t len, uint64_t flags)
 {
 	struct util_eq *eq;
 	struct util_event *entry;
 
 	eq = container_of(eq_fid, struct util_eq, eq_fid);
-	entry = malloc(sizeof(*entry) + len);
+	entry = calloc(1, sizeof(*entry) + len);
 	if (!entry)
 		return -FI_ENOMEM;
 
-	entry->size = (int) len;
+	entry->size = len;
 	entry->event = event;
 	entry->err = !!(flags & UTIL_FLAG_ERROR);
 	memcpy(entry->data, buf, len);
@@ -113,10 +159,12 @@ static ssize_t util_eq_write(struct fid_eq *eq_fid, uint32_t event,
 	return len;
 }
 
-static ssize_t util_eq_sread(struct fid_eq *eq_fid, uint32_t *event, void *buf,
-			     size_t len, int timeout, uint64_t flags)
+ssize_t ofi_eq_sread(struct fid_eq *eq_fid, uint32_t *event, void *buf,
+		     size_t len, int timeout, uint64_t flags)
 {
 	struct util_eq *eq;
+	uint64_t endtime;
+	ssize_t ret;
 
 	eq = container_of(eq_fid, struct util_eq, eq_fid);
 	if (!eq->internal_wait) {
@@ -124,18 +172,29 @@ static ssize_t util_eq_sread(struct fid_eq *eq_fid, uint32_t *event, void *buf,
 		return -FI_ENOSYS;
 	}
 
-	fi_wait(&eq->wait->wait_fid, timeout);
-	return fi_eq_read(eq_fid, event, buf, len, flags);
+	endtime = ofi_timeout_time(timeout);
+	do {
+		ret = fi_eq_read(eq_fid, event, buf, len, flags);
+		if (ret != -FI_EAGAIN)
+			break;
+
+		if (ofi_adjust_timeout(endtime, &timeout))
+			return -FI_EAGAIN;
+
+		ret = fi_wait(&eq->wait->wait_fid, timeout);
+	} while (!ret);
+
+	return ret == -FI_ETIMEDOUT ? -FI_EAGAIN : ret;
 }
 
-static const char *util_eq_strerror(struct fid_eq *eq_fid, int prov_errno,
-				    const void *err_data, char *buf, size_t len)
+const char *ofi_eq_strerror(struct fid_eq *eq_fid, int prov_errno,
+			    const void *err_data, char *buf, size_t len)
 {
-	return (buf && len) ? strncpy(buf, strerror(prov_errno), len) :
+	return (buf && len) ? strncpy(buf, fi_strerror(prov_errno), len) :
 			      fi_strerror(prov_errno);
 }
 
-static int util_eq_control(struct fid *fid, int command, void *arg)
+int ofi_eq_control(struct fid *fid, int command, void *arg)
 {
 	struct util_eq *eq;
 	int ret;
@@ -144,6 +203,7 @@ static int util_eq_control(struct fid *fid, int command, void *arg)
 
 	switch (command) {
 	case FI_GETWAIT:
+	case FI_GETWAITOBJ:
 		ret = fi_control(&eq->wait->wait_fid.fid, command, arg);
 		break;
 	default:
@@ -154,7 +214,7 @@ static int util_eq_control(struct fid *fid, int command, void *arg)
 	return ret;
 }
 
-static int util_eq_close(struct fid *fid)
+int ofi_eq_cleanup(struct fid *fid)
 {
 	struct util_eq *eq;
 	struct slist_entry *entry;
@@ -177,26 +237,41 @@ static int util_eq_close(struct fid *fid)
 			fi_close(&eq->wait->wait_fid.fid);
 	}
 
+	free(eq->saved_err_data);
 	fastlock_destroy(&eq->lock);
 	ofi_atomic_dec32(&eq->fabric->ref);
+	return 0;
+}
+
+static int util_eq_close(struct fid *fid)
+{
+	struct util_eq *eq;
+	int ret;
+
+	ret = ofi_eq_cleanup(fid);
+	if (ret)
+		return ret;
+
+	eq = container_of(fid, struct util_eq,
+			  eq_fid.fid);
 	free(eq);
 	return 0;
 }
 
 static struct fi_ops_eq util_eq_ops = {
 	.size = sizeof(struct fi_ops_eq),
-	.read = util_eq_read,
-	.readerr = util_eq_readerr,
-	.sread = util_eq_sread,
-	.write = util_eq_write,
-	.strerror = util_eq_strerror,
+	.read = ofi_eq_read,
+	.readerr = ofi_eq_readerr,
+	.sread = ofi_eq_sread,
+	.write = ofi_eq_write,
+	.strerror = ofi_eq_strerror,
 };
 
 static struct fi_ops util_eq_fi_ops = {
 	.size = sizeof(struct fi_ops),
 	.close = util_eq_close,
 	.bind = fi_no_bind,
-	.control = util_eq_control,
+	.control = ofi_eq_control,
 	.ops_open = fi_no_ops_open,
 };
 
@@ -216,7 +291,9 @@ static int util_eq_init(struct fid_fabric *fabric, struct util_eq *eq,
 		break;
 	case FI_WAIT_UNSPEC:
 	case FI_WAIT_FD:
+	case FI_WAIT_POLLFD:
 	case FI_WAIT_MUTEX_COND:
+	case FI_WAIT_YIELD:
 		memset(&wait_attr, 0, sizeof wait_attr);
 		wait_attr.wait_obj = attr->wait_obj;
 		eq->internal_wait = 1;
@@ -237,6 +314,51 @@ static int util_eq_init(struct fid_fabric *fabric, struct util_eq *eq,
 	return 0;
 }
 
+static int ofi_eq_match_fid_event(struct slist_entry *entry, const void *arg)
+{
+	fid_t fid = (fid_t) arg;
+	struct util_event *event;
+	struct fi_eq_cm_entry *cm_entry;
+	struct fi_eq_entry *cq_entry;
+
+	event = container_of(entry, struct util_event, entry);
+	cm_entry = (struct fi_eq_cm_entry *) event->data;
+
+	if (event->event == FI_CONNREQ &&
+	    fid == cm_entry->info->handle)
+		return 1;
+
+	cq_entry = (struct fi_eq_entry *)event->data;
+	return (fid == cq_entry->fid);
+}
+
+void ofi_eq_remove_fid_events(struct util_eq *eq, fid_t fid)
+{
+	struct fi_eq_err_entry *err_entry;
+	struct slist_entry *entry;
+	struct util_event *event;
+	struct fi_eq_cm_entry *cm_entry;
+
+	fastlock_acquire(&eq->lock);
+	while((entry =
+	      slist_remove_first_match(&eq->list, ofi_eq_match_fid_event,
+				       fid))) {
+		event = container_of(entry, struct util_event, entry);
+		if (event->err) {
+			err_entry = (struct fi_eq_err_entry *) event->data;
+			if (err_entry->err_data)
+				free(err_entry->err_data);
+
+		} else if (event->event == FI_CONNREQ) {
+			cm_entry = (struct fi_eq_cm_entry *) event->data;
+			assert(cm_entry->info);
+			fi_freeinfo(cm_entry->info);
+		}
+		free(event);
+	}
+	fastlock_release(&eq->lock);
+}
+
 static int util_verify_eq_attr(const struct fi_provider *prov,
 			       const struct fi_eq_attr *attr)
 {
@@ -244,7 +366,9 @@ static int util_verify_eq_attr(const struct fi_provider *prov,
 	case FI_WAIT_NONE:
 	case FI_WAIT_UNSPEC:
 	case FI_WAIT_FD:
+	case FI_WAIT_POLLFD:
 	case FI_WAIT_MUTEX_COND:
+	case FI_WAIT_YIELD:
 		break;
 	case FI_WAIT_SET:
 		if (!attr->wait_set) {
@@ -274,8 +398,8 @@ static int util_verify_eq_attr(const struct fi_provider *prov,
 	return 0;
 }
 
-int ofi_eq_create(struct fid_fabric *fabric_fid, struct fi_eq_attr *attr,
-		 struct fid_eq **eq_fid, void *context)
+int ofi_eq_init(struct fid_fabric *fabric_fid, struct fi_eq_attr *attr,
+		struct fid_eq *eq_fid, void *context)
 {
 	struct util_fabric *fabric;
 	struct util_eq *eq;
@@ -286,15 +410,11 @@ int ofi_eq_create(struct fid_fabric *fabric_fid, struct fi_eq_attr *attr,
 	if (ret)
 		return ret;
 
-	eq = calloc(1, sizeof(*eq));
-	if (!eq)
-		return -FI_ENOMEM;
-
+	eq = container_of(eq_fid, struct util_eq, eq_fid);
 	eq->fabric = fabric;
 	eq->prov = fabric->prov;
 	ret = util_eq_init(fabric_fid, eq, attr);
 	if (ret) {
-		free(eq);
 		return ret;
 	}
 
@@ -310,12 +430,29 @@ int ofi_eq_create(struct fid_fabric *fabric_fid, struct fi_eq_attr *attr,
 		ret = fi_poll_add(&eq->wait->pollset->poll_fid,
 				  &eq->eq_fid.fid, 0);
 		if (ret) {
-			util_eq_close(&eq->eq_fid.fid);
+			ofi_eq_cleanup(&eq->eq_fid.fid);
 			return ret;
 		}
 	}
 
-	*eq_fid = &eq->eq_fid;
 	return 0;
 }
 
+int ofi_eq_create(struct fid_fabric *fabric_fid, struct fi_eq_attr *attr,
+		  struct fid_eq **eq_fid, void *context)
+{
+	struct util_eq *eq;
+	int ret;
+
+	eq = calloc(1, sizeof(*eq));
+	if (!eq)
+		return -FI_ENOMEM;
+
+	ret = ofi_eq_init(fabric_fid, attr, &eq->eq_fid, context);
+	if (ret) {
+		free(eq);
+		return ret;
+	}
+	*eq_fid = &eq->eq_fid;
+	return 0;
+}
