@@ -36,8 +36,8 @@
 #include <errno.h>
 #include <sys/types.h>
 #include <stdlib.h>
-
-#include <fi_indexer.h>
+#include <assert.h>
+#include <ofi_indexer.h>
 
 /*
  * Indexer - to find a structure given an index
@@ -113,6 +113,31 @@ void *ofi_idx_remove(struct indexer *idx, int index)
 	return item;
 }
 
+void *ofi_idx_remove_ordered(struct indexer *idx, int index)
+{
+	struct ofi_idx_entry *entry;
+	void *item;
+	int temp_index;
+	int entry_index = ofi_idx_entry_index(index);
+
+	entry = idx->array[ofi_idx_array_index(index)];
+	item = entry[entry_index].item;
+	entry[entry_index].item = NULL;	
+	if (ofi_idx_free_list_empty(idx) || index < idx->free_list) {
+		entry[entry_index].next = idx->free_list;
+		idx->free_list = index;
+		return item;
+	}
+	temp_index = idx->free_list;
+	while (entry[ofi_idx_entry_index(temp_index)].next < index) {
+		temp_index = entry[ofi_idx_entry_index(temp_index)].next;
+	}
+	entry[entry_index].next = entry[ofi_idx_entry_index(temp_index)].next;
+	entry[ofi_idx_entry_index(temp_index)].next = index;
+
+	return item;
+}
+
 void ofi_idx_replace(struct indexer *idx, int index, void *item)
 {
 	struct ofi_idx_entry *entry;
